-- -*- coding: latin-1 -*-

--- Dieses Modul stellt Funktionen zum Zerlegen der Wortliste, zur
--- Prfung des Aufbaus der Wortliste und zur Prfung des Aufbaus von
--- Wrtern bereit.  Die API-Dokumentation kann mit <pre>
---
---   luadoc -d manual parse_wortliste.lua
---
--- </pre> erstellt werden.
--
-- @class module
-- @name parse_wortliste
-- @author Stephan Hennig
-- @copyright 2012, Stephan Hennig
--

--[[ just for luadoc 3.0 compatibility
module "parse_wortliste"
--]]


-- lokale Modul-Tabelle
local M = {}


-- Krzel fr LPEG-Funktionen.
local P = lpeg.P
local R = lpeg.R
local S = lpeg.S
local C = lpeg.C
local Cc = lpeg.Cc
local Cf = lpeg.Cf
local Ct = lpeg.Ct
local V = lpeg.V
-- Muster fr ein beliebiges Zeichen.
local any = P(1)



--
-- Beschreibung des Formats der Wortliste:
--
-- Jeder Datensatz (Zeile) entspricht einer bestimmten Schreibvariante
-- eines Wortes.  Unterschiedliche Schreibvarianten desselben Wortes
-- sind nicht miteinander verknpft.
--
-- Feldtrenner ist das Semikolon.
local sep = P";"
--
-- Kommentare werden durch '#' eingeleitet und erstrecken sich bis zum
-- Zeilenende.  Vor dem Kommentarzeichen sind beliebige Leerzeichen
-- erlaubt.
local com = P"#"
local spc = P" "
-- Muster fr ein Kommentar.
local opcomment = spc^0 * (com * any^0)^-1 * -1
-- Muster fr ein Kommentar mit Capture.  Die Capture enthlt den
-- Kommentartext ohne das Kommentarzeichen.
local opcommentC = spc^0 * (com * C(any^0))^-1 * -1
--
-- Leere Felder bestehen aus der Feldnummer eingeschlossen in
-- Minuszeichen, z. B. steht -2- fr ein leeres Feld 2.
--
-- Muster fr eine Ziffer.
local digit = R"09"
-- Muster fr ein beliebiges leeres Feld.
local leerX = P"-" * digit * P"-"
-- Muster fr leere Felder mit festem Inhalt.
local leer2 = P"-2-"
local leer3 = P"-3-"
local leer4 = P"-4-"
local leer5 = P"-5-"
local leer6 = P"-6-"
local leer7 = P"-7-"
local leer8 = P"-8-"
-- Krzel fr leere Felder mit voranstehendem Feldtrenner.
local _leer2 = sep * leer2
local _leer3 = sep * leer3
local _leer4 = sep * leer4
local _leer5 = sep * leer5
local _leer6 = sep * leer6
local _leer7 = sep * leer7
local _leer8 = sep * leer8
--
-- Belegte Felder bestehen aus beliebigen Zeichen auer Feldtrennern,
-- Leerzeichen oder Kommentarzeichen.  Eine przisere Beschreibung von
-- zulssigen Wrtern in Form einer Grammatik ist weiter unten zu
-- finden.
--
-- Muster fr ein Feld beliebigen Inhalts.
local feld = (any - (sep + spc + com))^1
-- Muster fr ein Feld beliebigen Inhalts mit Capture.  Die Capture
-- enthlt den Feldinhalt.
local feldC = C(feld)
-- Krzel fr ein Feld beliebigen Inhalts mit voranstehendem
-- Feldtrenner.
local _feldC = sep * feldC
-- Muster fr ein belegtes Feld.
local bfeld = feld - leerX
-- Krzel fr ein belegtes Feld mit voranstehendem Feldtrenner.
local _bfeld = sep * bfeld
--
-- Feld 1 enthlt ein Wort in ungetrennter Schreibung.
--
-- Die Felder 2, 3, 4 enthalten Wrter, die nicht ausschlielich in
-- Versalschreibung existieren.  Die Felder 3, 4 treten immer paarweise
-- auf.  Enthielten sie denselben Inhalt, so wird stattdessen Feld 2
-- verwendet.  Die Felder 3 und 4 existieren dann nicht.
--
-- Die Felder 5, 6, 7, 8 enthalten Wrter, die nur in expliziter
-- Versalschreibung existieren ('' durch 'ss' ersetzt).  Die Felder 6,
-- 7, 8 treten immer zu dritt auf.  Enthielten sie alle denselben
-- Inhalt, so wird stattdessen Feld 5 verwendet.  Die Felder 6, 7, 8
-- existieren dann nicht.
--
-- Feld     Beschreibung
--
--  1       * ungetrennt,
--
--  2       * keine explizite Versalschreibung,
--          * falls in allen Rechtschreibungen gleich,
--
--  3       * traditionelle Rechtschreibung,
--
--  4       * reformierte Rechtschreibung,
--
--  5       * explizite Versalschreibung,
--          * falls in allen Rechtschreibungen gleich,
--
--  6       * traditionelle Rechtschreibung in Deutschland und/oder
--            sterreich,
--
--  7       * reformierte Rechtschreibung,
--
--  8       * traditionelle Rechtschreibung in der Schweiz,
--



-- Tabelle mit Capture-Ersetzungen.  Bilde leere Felder auf den Wert
-- 'false' ab.
local _replace_empty_fields = {} for i = 1,8 do _replace_empty_fields["-"..i.."-"] = false end
-- Muster fr ein Feld mit Ersetzung.  Die Capture enthlt den Feldinhalt
-- oder 'false' fr leere Felder.
local feldRC = leerX / _replace_empty_fields + feldC
local _feldRC = sep * feldRC
-- Muster fr eine Zeile mit bis zu acht Feldern (mit Table-Capture) und
-- einem optionalen Kommentar.
local split_record = Ct(C(bfeld) * _feldRC^-7) * opcommentC
--
--- Zerlege einen Datensatz in einzelne Felder und speichere diese in
--- einer Tabelle.
--
-- @param record Datensatz
--
-- @return Tabelle, die einen Datensatz reprsentiert.  Nummerische
-- Indizes korrespondieren mit Feldnummern.  Leere Felder entsprechen
-- dem Wert <code>false</code>.  Der Schlssel 'comment' enthlt
-- gegebenenfalls einen Kommentar.
local function split(record)
   local t, comment = split_record:match(record)
   if comment then t.comment = comment end
   return t
end
M.split = split



--
-- Prfmuster fr Datensatzstruktur
--
-- Im folgenden werden unterschiedliche Typen von zulssigen Datenstzen
-- durch Zeichenketten reprsentiert.  Zeichenpositionen korrespondieren
-- dabei mit Feldernummern.  Das Zeichen an einer Position beschreibt,
-- ob das jeweilige Feld belegt ist oder leer.
--
-- Leere Felder werden durch '_' oder 'x' reprsentiert, wobei 'x' nur
-- aus Grnden der Lesbarkeit fr die Felder 2 und 5 verwendet wird.
--
-- Belegte Felder werden durch einen beschreibenden Buchstaben
-- gekennzeichnet.  Folgende Buchstaben werden verwendet:
--
-- Feld     Zeichen     Beschreibung
--
--  1        u          ungetrennt
--  2        a          alle
--  3        t          traditionelle Rechtschreibung
--  4        r          reformierte Rechtschreibung
--  5        c          Versalschreibung, alle
--  6        t          Versalschr., trad. Rechtschr. (D, AT)
--  7        r          Versalschr., reform. Rechtschr.
--  8        s          Versalschr., trad. Rechtschr. (CH)
--
-- Beispiele:
--
--   * Ein Wort, welches in allen Rechtschreibungen gleich getrennt
-- wird: Die Felder 1 und 2 sind belegt, gekennzeichnet durch die
-- Zeichen 'u' und 'a'.  Die Felder 3 bis 8 existieren nicht.  Typ:
-- 'ua'.
--
--   * Ein Wort, welches nicht nur in Versalschreibung existiert (also
-- ein Wort in normaler Schreibung), welches jedoch in traditioneller
-- und reformierter Rechtschreibung unterschiedlich getrennt wird: Die
-- Felder 1, 3 und 4 sind belegt ('u', 't', 'r').  Feld 2 ist leer
-- ('x').  Die Felder 5 bis 8 existieren nicht. Typ: 'uxtr'.
--
--   * Ein Wort, welches nur in Versalschreibung existiert und dort fr
-- alle Rechtschreibungen gleich getrennt wird: Die Felder 1 und 5 sind
-- belegt ('u' und 'c').  Die Felder 2 bis 4 sind leer ('x' und '_').
-- Die Felder 6 bis 8 existieren nicht.  Typ: 'ux__c'.
--
--   * Ein Wort, welches nur in (deutsch-)schweizerischer
-- Versalschreibung existiert: Die Felder 1 und 8 sind belegt ('u' und
-- 's').  Die Felder 2 bis 7 sind leer ('x' und '_').  Typ: 'ux__x__s'.
--
--
-- Diese Tabelle bildet Positionen (Indizes von 1-8) auf Tabellen
-- zulssiger Zeichen ab.  Die Untertabellen enthalten die an der
-- jeweiligen Position zulssigen Zeichen und bilden diese auf
-- Teilmuster ab.
local valid_flags = {
   [1] = { u = bfeld },-- ungetrennt
   [2] = { a = _bfeld, x = _leer2 },-- alle
   [3] = { t = _bfeld, _ = _leer3 },-- trad. RS
   [4] = { r = _bfeld, _ = _leer4 },-- reform. RS
   [5] = { c = _bfeld, x = _leer5 },-- Versalschr., alle
   [6] = { t = _bfeld, _ = _leer6 },-- Versalschr., trad. RS (D, AT)
   [7] = { r = _bfeld, _ = _leer7 },-- Versalschr., reform. RS
   [8] = { s = _bfeld, _ = _leer8 },-- Versalschr., trad. RS (CH)
}
--
-- Diese Variable enthlt ODER-verknpft (+) smtliche Muster, die
-- zulssige Datenstze reprsentieren.
local valid_records
--
--- <strong>nicht-ffentlich</strong> Fge der Liste aller Muster
--- zulssiger Datenstze ein neues Muster hinzu.  Die Captures der
--- Muster enthalten jeweils eine Zeichenkette, die den Datensatztyp
--- reprsentiert.
--
-- @param rec_type Zeichenkette, die ein Muster fr einen zulssigen
-- Datensatz reprsentiert
local function _add_to_valid_records(rec_type)
   local ch = string.sub(rec_type, 1, 1)
   local pat = valid_flags[1][ch]
   for i = 2,#rec_type do
      local ch = string.sub(rec_type, i, i)
      pat = pat * valid_flags[i][ch]
   end
   pat = Cc(rec_type) * pat * opcomment
   if valid_records then
      valid_records = valid_records + pat
   else
      valid_records = pat
   end
end
--
-- Erstelle ein Muster, welches smtliche zulssigen Datenstze
-- reprsentiert.
--
-- Muster fr Wrter, die nicht nur in Versalschreibung existieren.
-- (Die Felder 5 bis 8 existieren nicht.)
--
_add_to_valid_records("ua")
-- einfach;einfach
--
_add_to_valid_records("uxt_")
-- Abfallager;-2-;Abfa{ll/lll}ager;-4-
-- Abflu;-2-;Ab-flu;-4-
--
_add_to_valid_records("ux_r")
-- Abfalllager;-2-;-3-;Ab-fall=la-ger
--
_add_to_valid_records("uxtr")
-- abgelste;-2-;ab-ge-l-ste;ab-ge-ls-te
--
--
-- Muster fr Wrter, die nur in Versalschreibung existieren ('' durch
-- 'ss' ersetzt).
--
_add_to_valid_records("ux__c")
-- Abstoss;-2-;-3-;-4-;Abstoss
--
_add_to_valid_records("ux__xt__")
-- Litfasssulenstilleben;-2-;-3-;-4-;-5-;Lit-fass-su-len-sti{ll/ll-l}e-ben;-7-;-8-
--
_add_to_valid_records("ux__x_r_")
-- Fussballliga;-2-;-3-;-4-;-5-;-6-;Fussballliga;-8-
--
_add_to_valid_records("ux__x__s")
-- Litfassule;-2-;-3-;-4-;-5-;-6-;-7-;Litfa{ss/sss}ule
--
_add_to_valid_records("ux__xtr_")
-- ssssauer;-2-;-3-;-4-;-5-;ssssauer;ssssauer;-8-
--
_add_to_valid_records("ux__xt_s")
-- Fussballiga;-2-;-3-;-4-;-5-;Fussba{ll/lll}iga;-7-;Fussba{ll/lll}iga
--
_add_to_valid_records("ux__xtrs")
-- Fsse;-2-;-3-;-4-;-5-;Fsse;Fsse;Fsse
--
--
-- Muster fr Wrter, die in der reformierten Rechtschreibung
-- existieren, in der traditionellen Rechtschreibung jedoch nur in
-- Versalschreibweise ('' durch 'ss' ersetzt).
--
_add_to_valid_records("ux_rc")
-- Abfluss;-2-;-3-;Ab-fluss;Abfluss
--
_add_to_valid_records("ux_rxtr_")
-- Litfasssule;-2-;-3-;Litfasssule;-5-;Litfasssule;Litfasssule;-8-
--
_add_to_valid_records("ux_rxtrs")
-- dussligste;-2-;-3-;dussligste;-5-;dussligste;dussligste;dussligste
--
--
--- Ermittle den Typ eines Datensatzes.
--
-- @param record Datensatz
--
-- @return Zeichenkette, die den Datensatztyp reprsentiert.
local function identify_record(record)
   return valid_records:match(record)
end
M.identify_record = identify_record



--
-- Prfmuster fr Wortstruktur
--
local word_property = {}
--
--- <strong>nicht-ffentlich</strong> Lsche alle Worteigenschaften.
--
-- @return Tabelle <code>word_property</code>
local function _init_word_prop()
   word_property.has_invalid_alt = false
   word_property.has_nonstd = false
   word_property.has_nonstd_sss = false
   word_property.has_eszett = false
   return word_property
end
--
--- <strong>nicht-ffentlich</strong> Merke Eigenschaft
--- 'Spezialtrennung gefunden'.
--
-- @params c Capture
--
-- @return Capture
local function _word_prop_has_nonstd(c)
   word_property.has_nonstd = true
   return c
end
--
--- <strong>nicht-ffentlich</strong> Merke Eigenschaft
--- 'Dreikonsonantenregel fr s gefunden'.
--
-- @params c Capture
--
-- @return Capture
local function _word_prop_has_nonstd_sss(c)
   word_property.has_nonstd_sss = true
   return c
end
--
--- <strong>nicht-ffentlich</strong> Merke Eigenschaft
--- ' gefunden'.
--
-- @params c Capture
--
-- @return Capture
local function _word_prop_has_eszett(c)
   word_property.has_eszett = true
   return c
end
--
--- <strong>nicht-ffentlich</strong> Fge Strings zusammen.  Diese
--- Funktion akkumuliert Zeichen in einer Faltungscapture
--- (<code>lpeg.Cf</code>).
--
-- @param acc bisheriger Akkumulator
--
-- @param newvalue nchster Capturewert
--
-- @return neuer Akkumulatorzustand
local function _conc_cluster(acc, newvalue)
--   io.stderr:write("'",acc,"'..'",newvalue,"'\n")-- For debugging purposes.
   return acc..newvalue
end
--
--- <strong>nicht-ffentlich</strong> Vergleiche Strings.  Diese
--- Funktion wird in einer Faltungscapture (<code>lpeg.Cf</code>)
--- verwendet, um die Gleichheit zweier Captures festzustellen.  Falls
--- zwei Strings ungleich sind, so wird dies in Tabelle
--- <code>word_property</code> vermerkt.
--
-- @param a String 1
--
-- @param b String 2
--
-- @return String 1
local function _is_equal_part_word(a, b)
   word_property.has_invalid_alt = word_property.has_invalid_alt or (a ~= b)
   return a
end
--
-- Grammatik fr ein Wort.
--
-- Diese Grammatik prft nicht, ob die ersten und letzten beiden Zeichen
-- eines Wortes Buchstaben sind.  Die vollstndige Gltigkeit von
-- Wrtern sollte mit zustzlichen, nachgeschalteten Prfungen erfolgen.
--
local word = P{
   -- Initialregel.
   "word_with_property_table",
   --
   -- Wort
   --
   -- Whrend des Mustervergleichs werden verschiedene Eigenschaften des
   -- betrachteten Wortes in einer modullokalen Tabelle gespeichert.
   -- Die Flags innerhalb dieser Tabelle werden vor dem Mustervergleich
   -- gelscht.
   word_with_property_table = P"" / _init_word_prop * V"word" * -1,
   -- Ein Wort ist ein Teilwort, kann also fhrende oder abschlieende
   -- Trennzeichen enthalten.  Diese Generalisierung vereinfacht die
   -- Grammatik im Zusammenhang mit Alternativen etwas.
   word = V"part_word" * -1,
   -- Ein Teilwort besteht aus zusammenhngenden Zeichenketten,
   -- sogenannten Klustern, die durch Trennzeichen voneinander getrennt
   -- sind.  Ein Teilwort kann ein optionales fhrendes und schlieendes
   -- Trennzeichen enthalten.  Die Capture enthlt das Wort ohne
   -- jegliche Trennzeichen.
   part_word = V"ophyphen" * Cf(V"cluster" * (V"hyphen" * V"cluster")^0, _conc_cluster) * V"ophyphen",
   -- Ein Kluster besteht aus mehreren aneinandergereihten fundamentalen
   -- Klustern ohne Unterbrechung durch Trennzeichen.  Es gibt drei
   -- Arten von fundamentalen Klustern: Buchstabenkluster,
   -- Spezialtrennungen und Alternativen:
   --
   --                 B-cher   E{ck/k-k}e   Sto{ff/ff-f}et-zen   Wach[-s/s-]tu-be
   -- fund. Kluster   11-2222   1222222223   11122222222233-444   1111222222233-44
   -- Kluster         11-2222   1111111111   11111111111111-222   1111111111111-22
   --
   -- Die Capture enthlt smltliche Zeichen der aufeinanderfolgenden
   -- Kluster.
   cluster = Cf((V"cl_letter" + V"cl_nonstd" + V"cl_alt")^1, _conc_cluster),
   --
   -- Trennzeichen
   --
   -- Verschiedene Arten von Trennstellen werden durch unterschiedliche
   -- Trennzeichen markiert.
   --
   -- Die folgenden Trennzeichen werden verwendet:
   hyphen = S"=-\183.",
   -- Ein beliebiges, optionales Trennzeichen.
   ophyphen = V"hyphen"^-1,
   --
   -- Buchstabenkluster
   --
   -- Ein Buchstabenkluster besteht aus aufeinanderfolgenden Buchstaben.
   -- Die Capture sammelt alle Buchstaben.
   cl_letter = C(V"letter"^1),
   -- Die folgenden Buchstaben sind zulssig:
   letter = R("az", "AZ")
   + P"\196"-- 
   + P"\214"-- 
   + P"\220"-- 
   + P"\223" / _word_prop_has_eszett-- 
   + P"\224"-- 
   + P"\225"-- 
   + P"\226"-- 
   + P"\228"-- 
   + P"\231"-- 
   + P"\232"-- 
   + P"\233"-- 
   + P"\234"-- 
   + P"\235"-- 
   + P"\237"-- 
   + P"\238"-- 
   + P"\241"-- 
   + P"\243"-- 
   + P"\244"-- 
   + P"\246"-- 
   + P"\252"-- 
,
   --
   -- Spezialtrennung
   --
   -- Spezialtrennungen (non-standard hyphenation) beschreiben
   -- Trennregeln, die ber das bloe Einfgen eines Trennzeichens
   -- hinausgehen.  Dazu gehren die ck-Trennung und die
   -- Dreikonsonantenregel(n).  Spezialtrennungen sind jeweils
   -- eingeschlossen in Klammern.  Die Capture enthlt eine
   -- Zeichenkette, die der jeweils ungetrennten Spezialtrennung
   -- entspricht (z. B. 'ck' fr die ck-Trennung).
   cl_nonstd = V"nonstd_open" * V"nonstd_rule" * V"nonstd_close",
   -- Aufzhlung smtlicher Spezialtrennungen.  Das Auftreten von
   -- Spezialtrennungen wird in Tabelle <code>word_property</code>
   -- gespeichert.
   nonstd_rule = (V"ck" + V"fff" + V"lll" + V"mmm" + V"nnn" + V"ppp" + V"rrr" + V"ttt" + V"nonstd_sss") / _word_prop_has_nonstd,
   -- ck-Trennung: Die Capture enthlt die Zeichenfolge 'ck'.
   ck = C(P"ck") * V"nonstd_sep" * P"k" * V"hyphen" * P"k",
   -- Dreikonsonantenregel: Die Capture enthlt die Zeichenfolge fr die
   -- ungetrennte Konsonantenfolge.
   fff = C(P"ff") * V"nonstd_sep" * P"ff" * V"hyphen" * P"f",
   lll = C(P"ll") * V"nonstd_sep" * P"ll" * V"hyphen" * P"l",
   mmm = C(P"mm") * V"nonstd_sep" * P"mm" * V"hyphen" * P"m",
   nnn = C(P"nn") * V"nonstd_sep" * P"nn" * V"hyphen" * P"n",
   ppp = C(P"pp") * V"nonstd_sep" * P"pp" * V"hyphen" * P"p",
   rrr = C(P"rr") * V"nonstd_sep" * P"rr" * V"hyphen" * P"r",
   ttt = C(P"tt") * V"nonstd_sep" * P"tt" * V"hyphen" * P"t",
   sss = C(P"ss") * V"nonstd_sep" * P"ss" * V"hyphen" * P"s",
   -- Das Auftreten der Dreikonsonantenregel fr 's' wird in Tabelle
   -- <code>word_property</code> gespeichert.
   nonstd_sss = V"sss" / _word_prop_has_nonstd_sss,
   -- Klammer, die eine Spezialtrennung einfhrt.
   nonstd_open = P"{",
   -- Klammer, die eine Spezialtrennung abschliet.
   nonstd_close = P"}",
   -- Trennzeichen innerhalb einer Spezialtrennung.
   nonstd_sep = P"/",
   --
   -- Alternative
   --
   -- Eine Alternative beschreibt mehrdeutige Wortteile, die
   -- unterschiedlich getrennt werden knnen.  Die Capture enthlt die
   -- Zeichen der jeweils ersten Alternative ohne jegliche Trennzeichen.
   --
   -- Alternativen sind eingeschlossen in Klammern.
   cl_alt = V"alt_open" * V"alt_rule" * V"alt_close",
   -- Alternative Teilwrter werden durch ein spezielles Trennzeichen
   -- voneinander getrennt.
   alt_rule = Cf(V"alt_1st" * V"alt_sep" * V"alt_2nd", _is_equal_part_word),
   -- Die erste Alternative darf aus hchstens einem Kluster mit
   -- optionalem fhrenden und optionalem schlieenden Trennzeichen
   -- bestehen.  Trennzeichen innerhalb des Teilwortes sind nicht
   -- erlaubt.
   alt_1st = V"ophyphen" * V"cluster" * V"ophyphen",
   -- Die zweite Alternative entspricht einem Teilwort.  Dieses wird
   -- spter verworfen.
   alt_2nd = V"part_word",
   -- Klammer, die eine Alternative einfhrt.
   alt_open = P"[",
   -- Klammer, die eine Alternative abschliet.
   alt_close = P"]",
   -- Trennzeichen innerhalb einer Alternative.
   alt_sep = P"/",
}
--
--- Ermittle Eigenschaften eines Wortes.
--
-- @param rawword unbehandeltes Wort
--
-- @return <code>nil</code>, falls das Wort eine unzulssige Struktur
-- hat;</br> <code>props, parsed</code> sonst.  <code>props</code> ist
-- eine Tabelle mit Eigenschaften, <code>parsed</code> eine nicht nher
-- spezifizierte String-Capture des betrachteten Wortes.
local function parse_word(rawword)
   return word:match(rawword)
end
M.parse_word = parse_word



--- Normalisiere ein Wort.  Resultat ist ein Wort in einem fr Patgen
--- geeigneten Format, falls das Wort zulssig ist.  Smtliche
--- Trennzeichen werden durch '<code>-</code>' ersetzt.
--- Spezialtrennungen und Alternativen werden aufgelst.
--
-- @param rawword unbehandeltes Wort
--
-- @return <code>nil</code> fr Wrter mit unzulssiger Struktur;</br>
-- <code>word, props</code> sonst.  <code>word</code> ist das
-- normalisierte Wort, <code>props</code> ist eine Tabelle mit
-- Eigenschaften des betrachteten Wortes.
local function normalize_word(rawword)
   -- vorbereitende Prfung der Wortstruktur und Worteigenschaften
   local props = parse_word(rawword)
   -- Hat das Wort eine zulssige Struktur?
   if not props then return nil end
   -- Prfe auf unzulssige Alternativen.
   if props.has_invalid_alt then return nil end

   -- Ersetze Spezialtrennungen.
   rawword = string.gsub(rawword, "{(.-)/.-}", "%1")
   -- Ersetze Alternativen.
   rawword = string.gsub(rawword, "%[[-=\183%.]?(.-)[-=\183%.]?/.-%]", "%1")
   -- Ersetze Trennzeichen durch "-".
   rawword = string.gsub(rawword, "[=\183.]", "-")
   return rawword, props
end
M.normalize_word = normalize_word



--- Prfe ein Wort auf Wohlgeformtheit.
--
-- @param rawword Wort (normiert oder unbehandelt)
--
-- @return <code>nil</code> fr ein unzulssiges Wort;</br> <code>word,
-- props</code> sonst.  <code>word</code> ist das normalisierte Wort.
-- <code>props</code> ist eine Tabelle mit Eigenschaften des
-- betrachteten Wortes.
local function validate_word(rawword)
   -- Normalisiere Wort.
   local word, props = normalize_word(rawword)
   -- Zulssiges Wort?
   if not word then return nil end
   -- Prfe minimale Lnge.
   local len = string.len(word)
   if len < 2 then return nil end
   -- Prfe die fhrenden zwei Zeichen.
   for i = 1,2 do
      local ch = string.sub(word, i, i)
      if ch == "-" then return nil end
   end
   -- Prfe die letzten zwei Zeichen.
   for i = len-1,len do
      local ch = string.sub(word, i, i)
      if ch == "-" then return nil end
   end
   return word, props
end
M.validate_word = validate_word



--- Prfe einen Datensatz auf Wohlgeformtheit.  Geprft wird das Format
--- des Datensatzes und die Zulssigkeit smtlicher Wrter.
--
-- @param record zu prfender Datensatz
--
-- @return Typ des Datensatzes, falls dieser wohlgeformt ist;</br>
-- <code>nil</code>, falls der Datensatz kein zulssiges Format
-- hat;</br> <code>false</code>, falls der Datensatz unzulssige Wrter
-- enthlt.
local function validate_record(record)
   -- Zerlege Datensatz.
   local type = identify_record(record)
   -- Hat Datensatz zulssiges Format?
   if not type then return nil end
   -- Sind alle Wrter im Datensatz zulssig?
   local trec = split(record)
   -- Merke Inhalt von Feld 1 fr Konsistenzprfung des Datensatzes.
   local field1 = trec[1]
   if not field1 then return false end
   for i = 1,#trec do
      local word, props = trec[i]
      if word then
         -- Hat das Wort eine zulssige Struktur?
         word, props = validate_word(word)
         if not word then return false end
         -- Stimmt Wort mit Feld 1 berein?
         word = string.gsub(word, "-", "")
         if word ~= field1 then return false end
         -- Tritt eine Spezialtrennung an unzulssiger Feldnummer auf?
         if props.has_nonstd and (i==2 or i==4 or i==5 or i==7) then return false end
         -- Tritt eine Dreikonsonantenregel fr 's' an unzulssiger Feldnummer auf?
         if props.has_nonstd_sss and (i ~= 8) then return false end
         -- Tritt '' an unzulssiger Feldnummer auf?
         if props.has_eszett and (i > 4) then return false end
      end
   end
   return type
end
M.validate_record = validate_record



-- Exportiere Modul-Tabelle.
return M
